#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ant/globals.h>

#include "ttf.h"
#include "pixmap.h"
#include "max_rows.h"
#include "spline.h"

#define DEBUG_LINEARIZE 1
#define DEBUG 0

pixmapT *ttfRenderString(
    ttfHandle *handle	/* font handle */,
    const char *string	/* the string to render.
    			 * can include SGML style numeric  escapes.
                         */ ,
    int pixelsPerEm	/* Microsoft redefines the meaning of "em":
                         * Instead of the width of the letter "M" they
                         * use the distance between two baselines.
                         */,
    int platformID,
    int specificID
)
{
    pixmapT *bitmap;
    int last;


    int i, j;

    int charCode;
    int whichGlyph;
    ttfGlyph *glyph;
    ttfRenderList renderList[MAX_ROWS];
    int rows;
    int cols;
    int xoff;
    const char *p;

    /* determine size of bitmap */
    cols = 0;
    for (p = string; *p; p++) {
	if (*p == '&' && *(p + 1) == '#') {
	    /* SGML escape */
	    charCode = strtol(p + 2, (char **)&p, 10);
	    if (*p != ';') p--;
	} else {
	    charCode = (unsigned char) *p;
	}
	whichGlyph = ttfCode2Glyph(handle, charCode,
				   platformID, specificID);
	cols += handle->hmtxTable[whichGlyph].advanceWidth * pixelsPerEm
	/ handle->unitsPerEm;
    }
    rows = (handle->yMax - handle->yMin) * pixelsPerEm /
    handle->unitsPerEm + 1;
    assert (0 <= rows && rows < MAX_ROWS);
    bitmap = alloc_pixmap(rows, cols);

    xoff = 0;
    for (p = string; *p; p++) {
	/* The contour lists. Originally there were 16 contours of 1024
	 * points each. This proved to be too small, so I decided to make
	 * their size dynamic.
	 */
	pointlistT *Points;

	if (*p == '&' && *(p + 1) == '#') {
	    /* SGML escape */
	    charCode = strtol(p + 2, (char **)&p, 10);
	    if (*p != ';') p--;
	} else {
	    charCode = (unsigned char) *p;
	}
	whichGlyph = ttfCode2Glyph(handle, charCode, platformID, specificID);
	/* Read the specified glyph from the font */
	if ((glyph = ttfReadGlyph(handle, whichGlyph)) == NULL) {
	    fprintf(stderr, "readGlyph failed\n");
	    exit(1);
	}
	/* Convert each glyph contour to a polygon */
	last = 0;
	if (!(Points = malloc(sizeof(*Points) * glyph->nContours))) {
	    fprintf(stderr, "%s: cannot malloc space for %d contours: %s\n",
	    		    cmnd, glyph->nContours, strerror(errno));
	    exit(1);
	}
	for (i = 0; i < glyph->nContours; i++) {
	    ttfLinearize(&Points[i],
			  glyph->endPoints[i] - last + 1,
			  glyph->points + last,
			  pixelsPerEm,
			  handle->unitsPerEm);
            #if DEBUG & DEBUG_LINEARIZE
	    {
		int j;
		for (j = 0; j < Points[i].u; j++) {
		    fprintf(stderr, "point %d.%d: (%d, %d)\n",
		    	    i, j,
			    Points[i].p[j].x, Points[i].p[j].y);
		}
	    }
	    #endif
	    last = glyph->endPoints[i] + 1;
	}

	/* Move everything about a bit to fit in a grid with origo (0,0) */
	for (i = 0; i < glyph->nContours; i++)
	    for (j = 0; j < Points[i].u; j++) {
		Points[i].p[j].x = Points[i].p[j].x;	/* XXX ??? */
		Points[i].p[j].y = handle->yMax - Points[i].p[j].y;
	    }

	/* Scan-convert the polygon to a render list (actually a scan-line list */
	memset(renderList, 0, sizeof(renderList));
	for (i = 0; i < glyph->nContours; i++)
	    ttfRender(renderList, &Points[i],
	    	      pixelsPerEm, handle->unitsPerEm);

	/* free the contour lists */
	for (i = 0; i < glyph->nContours; i++) {
	    free(Points[i].p);
	    Points[i].a = Points[i].u = 0;
	}
	free(Points);

	/* Rasterize the render-list into a list of pixels to turn on */
	ttfRasterize(bitmap, xoff, renderList,
		     pixelsPerEm, handle->unitsPerEm);
	xoff += handle->hmtxTable[whichGlyph].advanceWidth * pixelsPerEm
	/ handle->unitsPerEm;
    }
    return bitmap;
}


/*
 * return a map of valid chars in the font.
 * Each character is represented by 1 char. 
 * Bit 0 is set if the character is present
 * The last character has bit 1 set.
 */
char *ttfCharMap(
    ttfHandle *handle,
    int platformID,
    int specificID
){
    char *charmap = malloc(65536);
    int i;
    int lastchar = 0;

    for (i = 0; i < 65536; i++) {
	int glyph = ttfCode2Glyph(handle, i, platformID, specificID);
	if (glyph == -1) {
	    charmap[i] = 0;
	} else {
	    charmap[i] = 1;
	    lastchar = i;
	}
    }
    charmap[lastchar] |= 2;

    return charmap;
    
}
